/** ZCT/init.c--Created 060205 **/
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "smp.h"

/**
initialize_data():
Sets up all of the arrays needed for the program.
Created 070305; last modified 011808
**/
void initialize_data(void)
{
	int x;
	int y;
	int z;
	BITBOARD mask;
	COLOR c;
	PIECE p;
	SQUARE s;
	BITBOARD (*fill_dir[8])(BITBOARD pieces, BITBOARD empty) =
	{
		fill_up_left, fill_up, fill_up_right, fill_right, fill_down_right, fill_down, fill_down_left, fill_left
	};

	/* Initialize the standard global stuff. */
	protocol = DEFAULT;

	input_stream = stdin;
	output_stream = stdout;
	log_stream = fopen("logfile.txt", "at+");

	setbuf(input_stream, NULL);
	setbuf(output_stream, NULL);
	setbuf(log_stream, NULL);

	fprintf(output_stream, "ZCT Version %s\n\n", ZCT_VERSION);
	fprintf(log_stream, "ZCT Version %s\n\n", ZCT_VERSION);

	post = TRUE;
	hard = TRUE;
	set_time_control(0, 5 * 60000, 0);
	nodes_per_check = 10000;

	/* attack bitboards */
	for (s = A1; s < OFF_BOARD; s++)
		mask_bb[s] = (BITBOARD)1 << s;
	for (s = A1; s < OFF_BOARD; s++)
	{
		mask = MASK(s);
		knight_moves_bb[s] = SHIFT_UP(SHIFT_UL(mask)) |
			SHIFT_UP(SHIFT_UR(mask)) |
			SHIFT_DN(SHIFT_DL(mask)) |
			SHIFT_DN(SHIFT_DR(mask)) |
			SHIFT_LF(SHIFT_UL(mask)) |
			SHIFT_LF(SHIFT_DL(mask)) |
			SHIFT_RT(SHIFT_UR(mask)) |
			SHIFT_RT(SHIFT_DR(mask));
		king_moves_bb[s] = SHIFT_UL(mask) | SHIFT_UP(mask) |
			SHIFT_UR(mask) | SHIFT_RT(mask) | SHIFT_DR(mask) |
			SHIFT_DN(mask) | SHIFT_DL(mask) | SHIFT_LF(mask);
		for (x = A1; x < OFF_BOARD; x++)
		{
			mask = (BITBOARD)0;
			/* Inefficient: try all directions and only if they match up set the bitboard. */
			for (y = 0; y < 8; y++)
				mask |= fill_dir[y](MASK(s), ~(BITBOARD)0) & fill_dir[(y + 4) & 7](MASK(x), ~(BITBOARD)0);
			in_between_bb[s][x] = mask & ~MASK(s) & ~MASK(x);
		}
		for (c = WHITE; c <= BLACK; c++)
		{
			mask = MASK(s);
			mask = SHIFT_FORWARD(mask, c);
			mask = SHIFT_LF(mask) | SHIFT_RT(mask);
			pawn_caps_bb[c][s] = mask;
		}
	}

	castle_ks_mask[WHITE] = MASK(F1) | MASK(G1);
	castle_ks_mask[BLACK] = MASK(F8) | MASK(G8);
	castle_qs_mask[WHITE] = MASK(D1) | MASK(C1) | MASK(B1);
	castle_qs_mask[BLACK] = MASK(D8) | MASK(C8) | MASK(B8);

	/* zobrist arrays for hashkeys */
	for (x = 0; x < 1000; x++)
		random_hashkey();
	for (c = WHITE; c <= BLACK; c++)
	{
		for (p = PAWN; p <= KING; p++)
		{
			for (s = A1; s < OFF_BOARD; s++)
			{
				zobrist_piece[c][p][s] = random_hashkey();
				/* Check against other keys to ensure good randoms. */
				for (x = WHITE; x <= c; x++)
					for (y = PAWN; y <= p; y++)
						for (z = A1; z < s; z++)
							if (zobrist_piece[c][p][s] == zobrist_piece[x][y][z])
								zobrist_piece[c][p][s] ^= random_hashkey() ^ random_hashkey();
			}
		}
	}
	for (s = A1; s < OFF_BOARD; s++)
		zobrist_ep[s] = random_hashkey();
	
	for (x = 0; x < 16; x++)
		zobrist_castle[x] = random_hashkey();

	zobrist_stm = random_hashkey();

	/* If we are using SMP, we need to allocate the hash table in shared memory, done elsewhere.
	   The other tables are allocated one per process. */
	hash_alloc(hash_size);
	qsearch_hash_table = (HASH_ENTRY *)calloc(qsearch_hash_size, sizeof(HASH_ENTRY));
	pawn_hash_table = (PAWN_HASH_ENTRY *)calloc(pawn_hash_size, sizeof(PAWN_HASH_ENTRY));
	eval_hash_table = (EVAL_HASH_ENTRY *)calloc(eval_hash_size, sizeof(EVAL_HASH_ENTRY));
}

/**
initialize_board():
Initializes board to the starting position if given a NULL argument, or the given FEN otherwise.
Created 060205; last modified 092507
**/
void initialize_board(char *fen)
{
	const char castle_string[5] = "KkQq";
	enum { INIT_BOARD, INIT_STM, INIT_CASTLE, INIT_EP, INIT_FIFTY, INIT_MOVE_COUNT } init_state;
	char *p;
	char *r;
	int x;
	COLOR c;
	SQ_FILE file;
	SQ_RANK rank;
	SQUARE square;

	game_over = FALSE;
	game_result = INCOMPLETE;
	search_age = 0;

	if (fen == NULL)
	{
		fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";
		zct_side = BLACK;
	}
	else
		zct_side = EMPTY;

	/* Initialize. */
	for (square = A1; square < OFF_BOARD; square++)
	{
		board.piece[square] = EMPTY;
		board.color[square] = EMPTY;
	}
	board.castle_rights = CASTLE_NO;
	board.ep_square = OFF_BOARD;
	board.fifty_count = 0;
	board.move_number = 1;

	square = A8;
	file = FILE_A;
	rank = RANK_8;
	init_state = INIT_BOARD;

	/* Start reading the board information from the FEN. */
	for (p = fen; *p != '\0'; p++)
	{
		if (*p == ' ')
		{
			init_state++;
			continue;
		}
		switch (init_state)
		{
			case INIT_BOARD:
				if (*p == '/')
				{
					rank--;
					file = FILE_A;
				}
				if (rank < RANK_1 || file > FILE_H)
				{
					initialize_board(NULL);
					print("FEN parse error: square off board\n");
					return;
				}
				square = SQ_FROM_RF(rank, file);
				for (c = WHITE; c <= BLACK; c++)
				{
					if ((r = strchr(piece_char[c], *p)) != NULL)
					{
						board.piece[square] = r - piece_char[c]; 
						board.color[square] = c;
						square++;
						file++;
						break;
					}
				}
				if (*p >= '1' && *p <= '8')
				{
					for (x = *p - '0'; x >= 1; x--)
					{
						board.piece[square] = EMPTY;
						board.color[square] = EMPTY;
						square++;
						file++;
					}
				}
				break;
			case INIT_STM:	
				if (*p == 'W' || *p == 'w')
					board.side_tm = WHITE;
				else if (*p == 'B' || *p == 'b')
					board.side_tm = BLACK;
				else
				{
					initialize_board(NULL);
					print("FEN parse error: bad side to move char '%c'\n", *p);
					return;
				}
				board.side_ntm = COLOR_FLIP(board.side_tm);
				break;
			case INIT_CASTLE:
				if ((r = strchr(castle_string, *p)) != NULL)
					board.castle_rights |= 1 << (r - castle_string);
				else if (*p != '-')
				{
					initialize_board(NULL);
					print("FEN parse error: bad castle status char '%c'\n", *p);
					return;
				}
				break;
			case INIT_EP:
				if (*p >= 'a' && *p <= 'h' && *(p + 1) >= '1' && *(p + 1) <= '8')
				{
					board.ep_square = SQ_FROM_RF(*(p + 1) - '1', *p - 'a');
					p++;
				}
				else if (*p != '-')
				{
					initialize_board(NULL);
					print("FEN parse error: bad ep square '%c%c'\n", *p, *(p + 1));
					return;
				}
				break;
			case INIT_FIFTY:
				board.fifty_count = atoi(p);
				while (*p >= '0' && *p <= '9')
					p++;
				p--;
				break;
			case INIT_MOVE_COUNT:
				board.move_number = atoi(p);
				while (*p >= '0' && *p <= '9')
					p++;
				p--;
				break;
			default:
				continue;
		}
	}
	if (init_state < INIT_CASTLE)
	{
		initialize_board(NULL);
		print("FEN parse error: not enough information\n");
		return;
	}
	initialize_bitboards();
	initialize_hashkey();
	reset_clocks();

	board.game_entry = &board.game_stack[0];
}

/**
initialize_bitboards():
Set up piece and occupied bitboards from the piece[] and color[] arrays. Also sets up
other board representation data.
Created 070105; last modified 102806
**/
void initialize_bitboards(void)
{
	PIECE p;
	SQUARE square;

	board.color_bb[WHITE] = (BITBOARD)0;
	board.color_bb[BLACK] = (BITBOARD)0;
	board.occupied_bb = (BITBOARD)0;
	for (p = 0; p < 6; p++)
		board.piece_bb[p] = (BITBOARD)0;
	board.piece_count[WHITE] = 0;
	board.piece_count[BLACK] = 0;

	for (square = A1; square < OFF_BOARD; square++)
	{
		if (board.color[square] != EMPTY)
		{
			SET_BIT(board.occupied_bb, square);
			SET_BIT(board.color_bb[board.color[square]], square);
			SET_BIT(board.piece_bb[board.piece[square]], square);
			if (board.piece[square] != PAWN)
				board.piece_count[board.color[square]]++;
		}
	}

	board.king_square[WHITE] = first_square(board.piece_bb[KING] & board.color_bb[WHITE]);
	board.king_square[BLACK] = first_square(board.piece_bb[KING] & board.color_bb[BLACK]);
}

/**
initialize_hashkey():
Initializes the main hashkey and pawn hashkey with the zobrist arrays.
Created 080906; last modified 083106
**/
void initialize_hashkey(void)
{
	BITBOARD pieces;
	SQUARE square;

	board.hashkey = (HASHKEY)0;
	board.pawn_hash_entry.hashkey = (HASHKEY)0;
	pieces = board.occupied_bb;
	while (pieces)
	{
		square = first_square(pieces);
		CLEAR_BIT(pieces, square);
		
		board.hashkey ^= zobrist_piece[board.color[square]][board.piece[square]][square];
		if (board.piece[square] == PAWN)
			board.pawn_hash_entry.hashkey ^= zobrist_piece[board.color[square]][PAWN][square];
	}
	if (board.side_tm == BLACK)
		board.hashkey ^= zobrist_stm;
	board.hashkey ^= zobrist_castle[board.castle_rights];
	board.hashkey ^= zobrist_ep[board.ep_square];
}

/**
hash_alloc():
Allocate the hash table in either shared or local memory, depending on whether we are using SMP.
Note that the argument is the number of entries, not the size in bytes.
Created 123107; last modified 123107
**/
void hash_alloc(int hash_table_size)
{
#ifdef SMP
	if (hash_table != NULL)
		munmap(hash_table, hash_size * sizeof(HASH_ENTRY));
	hash_table = (HASH_ENTRY *)shared_alloc(hash_table_size * sizeof(HASH_ENTRY));
#else
	if (hash_table != NULL)
		free(hash_table);
	if ((hash_table = (HASH_ENTRY *)calloc(hash_table_size, sizeof(HASH_ENTRY))) == NULL)
	{
		perror("calloc");
		exit(0);
	}
#endif
	hash_size = hash_table_size;
}
