/** ZCT/smp.h--Created 081305 **/
#include "zct.h"

#ifndef SMP_H
#define SMP_H

#ifdef SMP

#include <fcntl.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/mman.h>
#include <sys/sem.h>
#include <signal.h>

//#define DEBUG_SMP

#ifdef DEBUG_SMP
#	define SMP_DEBUG(x)			x
#else
#	define SMP_DEBUG(x)
#endif

#define MAX_SPLIT_POINTS		(MAX_CPUS * MAX_CPUS)
#define MAX_MESSAGES			(128)

#ifdef ZCT_OSX

#	include <libkern/OSAtomic.h>

typedef OSSpinLock LOCK_T[1];

#	define LOCK(l)				(OSSpinLockLock(l))
#	define UNLOCK(l)			(OSSpinLockUnlock(l))
#	define LOCK_INIT(l)			((l)[0] = 0)
#	define LOCK_FREE(l)			((l)[0] = 0)

#elif defined(ZCT_x86)

#	ifdef ZCT_POSIX

typedef volatile int LOCK_T[1];

/* From Crafty. Bob says he took it from the Linux kernel. */
static void __inline__ LOCK(volatile int *lock)
{
	int dummy;
	asm __volatile__ (
		"1:          movl    $1, %0" "\n\t"
		"            xchgl   (%1), %0" "\n\t"
		"            testl   %0, %0" "\n\t"
		"            jz      3f" "\n\t"
		"2:          pause" "\n\t"
		"            movl    (%1), %0" "\n\t"
		"            testl   %0, %0" "\n\t"
		"            jnz     2b" "\n\t"
		"            jmp     1b" "\n\t"
		"3:" "\n\t"
		:"=&q" (dummy)
		:"q" (lock)
		:"cc", "memory");
}

#	elif defined(ZCT_WINDOWS)

typedef volatile long LOCK_T[1];

#		define LOCK(l) while (InterlockedExchange((l),1) != 0) while ((l)[0] == 1)

#	endif

#	define UNLOCK(l)			((l)[0] = 0)
#	define LOCK_INIT(l)			((l)[0] = 0)
#	define LOCK_FREE(l)			((l)[0] = 0)

#else /* !ZCT_x86 */

#	define ZCT_USE_SEMAPHORES

typedef volatile int LOCK_T;

#	define LOCK(l)				(lock(l))
#	define UNLOCK(l)			(unlock(l))
#	define LOCK_INIT(l)
#	define LOCK_FREE(l)

#endif

/* These are messages passed around by processors to coordinate DTS searching. Look for the functions that implement
	their actions to get a better idea of what each command does. */
typedef enum { SMP_INIT = 1, SMP_SEARCH, SMP_SPLIT, SMP_PERFT, SMP_GO } SMP_INPUT;
typedef enum { SMP_DONE = 1 } SMP_OUTPUT;
/* These are asynchronous commands, meaning that the sending processor does not wait for a reply. */
typedef enum { SMP_HELP = 1, SMP_UPDATE, SMP_INSERT, SMP_UNSPLIT, SMP_STOP, SMP_PARK } SMP_MESSAGE_TYPE;

typedef struct
{
	SMP_MESSAGE_TYPE message;
	int data;
} SMP_MESSAGE;

typedef struct SPLIT_POINT
{
	int id;
	volatile BOOL active;
	volatile BOOL no_moves_left;
	BOARD board;
	MOVE move_path[GAME_STACK_SIZE];
	MOVE *last_move;
	volatile int child_count;
	volatile BOOL update[MAX_CPUS];
	volatile BOOL insert[MAX_CPUS];
	volatile BOOL is_child[MAX_CPUS];
	volatile MOVE move_list[256];
	volatile MOVE pv[MAX_PLY];
	SEARCH_BLOCK * volatile sb;
	LOCK_T lock; /* Used for all data in each split point */
} SPLIT_POINT;

typedef struct /* Wrapper struct for each process, used for process communication */
{
	int id;
	int pid;
	BITBOARD nodes;
	BITBOARD q_nodes;
	int idle_time;
	int last_idle_time;
	BOOL idle;
	SMP_MESSAGE message_queue[MAX_MESSAGES];
	volatile int message_count;
	/* The input and output variables are grouped with the data. Any message that is sent to another
		process can be accompanied by a word of data. */
	volatile SMP_INPUT input;
	volatile SMP_OUTPUT output;
	volatile int data;
	LOCK_T lock;
	LOCK_T input_lock; /* Used for all input/output */
} SMP_BLOCK;

typedef struct /* Wrapper struct for global smp data */
{
	int splits_done;
	int stops_done;
	BOARD root_board;
	int split_id;
	int split_count;
	BOOL stop;
	BOOL return_flag;
	int current_depth;
	int return_value;
	MOVE return_pv[MAX_PLY];
	LOCK_T lock; /* Used for general smp data, split points, etc. */
	LOCK_T io_lock; /* Used for all input/output */
} SMP_DATA;

typedef struct
{
	volatile BOOL used;
	SEARCH_BLOCK search_stack[MAX_PLY];
	SEARCH_BLOCK *search_block;
} TREE_BLOCK;

extern SMP_BLOCK *smp_block;
extern SPLIT_POINT *split_point;
extern SMP_DATA *smp_data;
extern TREE_BLOCK *tree_block;

/* smp.c */
void *shared_alloc(BITBOARD size);
void idle_loop(int id);
int smp_tell(int id, SMP_INPUT input, int data);
void smp_message(int id, SMP_MESSAGE_TYPE message, int data);
void smp_copy_to(SPLIT_POINT *to, GAME_ENTRY *ge, SEARCH_BLOCK *sb);
void smp_copy_from(SPLIT_POINT *from, SEARCH_BLOCK *sb);
void smp_copy_root(BOARD *to, BOARD *from);
void move_list_get(SEARCH_BLOCK *sb);
void move_list_put(SEARCH_BLOCK *sb);
void handle_messages(SEARCH_BLOCK **sb);
void start_child_processors(void);
void stop_child_processors(void);
#ifdef ZCT_USE_SEMAPHORES
void lock(LOCK_T l);
void unlock(LOCK_T l);
#endif
void smp_cleanup(void);
void smp_cleanup_sig(int x);
/* smp2.c */
int split(SEARCH_BLOCK *sb, int ply, int id);
SEARCH_BLOCK *attach(int sp);
void merge(SEARCH_BLOCK *sb);
SEARCH_BLOCK *update(SEARCH_BLOCK *sb, int id);
void insert(BOARD *current_board, int id, SPLIT_POINT *new_sp);
SEARCH_BLOCK *detach(SEARCH_BLOCK *sb);
SEARCH_BLOCK *stop(SEARCH_BLOCK *sb);
SEARCH_BLOCK *unsplit(SEARCH_BLOCK *sb, int id);
void copy_search_state(SEARCH_BLOCK *sb, int process);
int find_split_point(void);

#else

#define LOCK(x)
#define UNLOCK(x)

#endif /* SMP */

#endif /* SMP_H */
