/** ZCT/smp2.c--Created 102806 **/
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "smp.h"
#include <math.h>

#ifdef SMP

/**
split():
In the tree, sets up the data for using multiple processors on a single node.
Created 081706; last modified 022308
**/
int split(SEARCH_BLOCK *sb, int ply, int id)
{
	int sp;
	int process;
	int *s_ply;
	MOVE *from;
	MOVE *to;
	SEARCH_BLOCK *old_sb;
	SPLIT_POINT **s_point;
	SPLIT_POINT *last_split_point;
	GAME_ENTRY *current;

	old_sb = sb;
	last_split_point = NULL;
	if (*board.split_ply > ply)
		return -1;
	else if (*board.split_ply == ply)
		return *board.split_point - split_point;
	/* Make sure there isn't already a split point at this ply, which can often happen at the
		beginning of an iteration with more than 2 processors, and each child selects the same node.
		Also keep track of any split point at a node below where we are trying to split. */
	/*for (s_ply = board.split_ply, s_point = board.split_point; *s_ply != -1 && *s_ply >= ply;
		last_split_point = *s_point, s_ply--, s_point--)
	{
		if (*s_ply == ply)
			return *s_point - split_point;
	}*/
	/* This is obvious. :) */
	if (sb->ply < ply)
		return -1;
	if (*board.split_point != NULL && (*board.split_point)->sb->ply >= ply)
		return -1;

	LOCK(smp_data->lock);
	/* Check for overflow of split_point[] or board.split_point_stack[]. */
	if (smp_data->split_count == MAX_SPLIT_POINTS || board.split_point - board.split_point_stack == MAX_PLY)
	{
		UNLOCK(smp_data->lock);
		return -1;
	}
	/* Find a split point. */
	for (sp = 0; sp < MAX_SPLIT_POINTS; sp++)
		if (!split_point[sp].active)
			break;
	ASSERT(sp != MAX_SPLIT_POINTS);
	/* Now unmake moves until we are at the position of the new split point. */
	current = board.game_entry;
	ASSERT(ply > 0 && ply < MAX_PLY);
	while (sb->ply >= ply)
	{
		/* Check if we have made a move that needs to be unmade at this ply. Certain SEARCH_STATEs
			indicate that a move has not yet been made. Unless this is the top ply, only IID will be found.
			Note that we only go back in the game entries, because we copy the move list up to that point for the split. */
		if (sb->search_state != SEARCH_START && sb->search_state != QSEARCH_START && sb->search_state != SEARCH_IID)
			current--;
		sb--;
	}
	sb++;
	ASSERT(sb->ply == ply);
	/* Make sure this is the intended split point. */
	if (sb->id * MAX_PLY != id * MAX_PLY) /* HACKY! */
	{
		UNLOCK(smp_data->lock);
		/* Remake those moves we just unmade... */
		while (current < board.game_entry)
			current++;
		return -1;
	}
	/* We found a split point. Start to split. */
	LOCK(split_point[sp].lock);
	/* Each split point gets a unique id to handle asynchronous messages. We need to be able to handle
		messages from a processor at a split point that the current processor has already left. */
	split_point[sp].id = smp_data->split_id++;
	SMP_DEBUG(print("cpu %i splitting, sp=%i id=%i %E\n", board.id, sp, split_point[sp].id, sb));
	/* Set the data in BOARD to tell where this split point is. */
	/* Find where to insert this split point. */
	if (last_split_point == NULL)
	{
		*++board.split_ply = ply;
		*++board.split_point = &split_point[sp];
	}
	else
	{
		/* We are "inserting" the split point, so shift all others down. */
		insert(&board, last_split_point->id, &split_point[sp]);
	}
	smp_copy_to(&split_point[sp], current, sb);
	split_point[sp].sb = sb - board.search_stack + split_point[sp].board.search_stack;
	split_point[sp].sb->search_state = SEARCH_JOIN;
	split_point[sp].sb->move_made = FALSE;
	split_point[sp].pv[0] = NULL_MOVE;
	/* Copy the move list into shared memory */
	to = split_point[sp].move_list;
	from = sb->next_move;
	while (from < sb->last_move)
		*to++ = *from++;
	*to = NULL_MOVE;
	/* We only change this after the split point is ready. */
	split_point[sp].is_child[board.id] = TRUE;
	split_point[sp].child_count = 1;
	split_point[sp].no_moves_left = FALSE;
	split_point[sp].active = TRUE;
	UNLOCK(split_point[sp].lock);

	smp_data->split_count++;
	smp_data->splits_done++;
	UNLOCK(smp_data->lock);

	if (last_split_point != NULL)
	{
		/* Tell any processors to insert this split point if there's a split point above it. */
		LOCK(last_split_point->lock);
		for (process = 0; process < process_count; process++)
		{
			if (process != board.id && last_split_point->is_child[process])
			{
				last_split_point->insert[process] = TRUE;
				smp_message(process, SMP_INSERT, last_split_point->id);
			}
		}
		UNLOCK(last_split_point->lock);
	}
	/* Now go back to the position we were searching. */
	while (current < board.game_entry)
		current++;
	sb = old_sb;
	return sp;
}

/**
attach():
This attaches an idle process onto an existing split point. It returns if the split point
is already dead.
Created 082806; last modified 022308
**/
SEARCH_BLOCK *attach(int sp)
{
	SPLIT_POINT **s;
//	SPLIT_POINT *sp_stack[MAX_PLY + 1];

	LOCK(split_point[sp].lock);
	if (!split_point[sp].active || split_point[sp].no_moves_left)
	{
		UNLOCK(split_point[sp].lock);
		SMP_DEBUG(print("cpu %i failed attaching, sp %i no longer active...\n", board.id, sp));
		return NULL;
	}
	/* Lock and check all split points under this one as well, as we need to attach "atomically". */
	for (s = split_point[sp].board.split_point - 1; *s != NULL; s--)
	{
		LOCK((*s)->lock);
		if (!(*s)->active || (*s)->no_moves_left)
		{
			for (; s < split_point[sp].board.split_point; s++)
				UNLOCK((*s)->lock);
			UNLOCK(split_point[sp].lock);
			SMP_DEBUG(print("cpu %i failed attaching, sp %i no longer active...\n", board.id, sp));
			return NULL;
		}
	}
	/* Copy the board from the split point to our local board. Then attach to all the split points
		below this one. Note that all of them are still locked. */
	smp_copy_from(&split_point[sp], split_point[sp].sb);
	for (s = board.split_point_stack + 1; s < board.split_point; s++)
	{
		(*s)->child_count++;
		(*s)->is_child[board.id] = TRUE;
		SMP_DEBUG(print("cpu %i attaching to %i, now %i children and %i splitpoints.\n%B", board.id, split_point[sp].id,
			split_point[sp].child_count, board.split_point - board.split_point_stack, &board));
		UNLOCK((*s)->lock);
	}
	/* Update the split point. */
	split_point[sp].child_count++;
	split_point[sp].is_child[board.id] = TRUE;
	SMP_DEBUG(print("cpu %i attaching to %i, now %i children and %i splitpoints.\n%B", board.id, split_point[sp].id,
		split_point[sp].child_count, board.split_point - board.split_point_stack, &board));
	UNLOCK(split_point[sp].lock);
	/* We now take away our idle status.  */
	smp_block[board.id].idle = FALSE;
	smp_block[board.id].idle_time += get_time() - smp_block[board.id].last_idle_time;
	smp_block[board.id].last_idle_time = 0;

	return split_point[sp].sb - split_point[sp].board.search_stack + board.search_stack;
}

/**
merge():
This function is called when we get a value > alpha at a split point, meaning
that other processors might be doing unnecessary work, and need to check.
Created 082906; last modified 041807
**/
void merge(SEARCH_BLOCK *sb)
{
	int x;
	SPLIT_POINT *sp;

	sp = *board.split_point;
	LOCK(sp->lock);
	/* Make sure this value is still the best. */
	if (sb->best_score >= sp->sb->alpha)
	{
		sp->sb->alpha = sb->best_score;
		sp->sb->pv_found = sb->pv_found;
		copy_pv(sp->pv, board.pv_stack[sb->ply]);
		SMP_DEBUG(print("cpu %i merging score=%V pv=%lM %E\n", board.id, sb->best_score, board.pv_stack[sb->ply], sb));
		/* Tell the other processors to update. */
		for (x = 0; x < process_count; x++)
			if (x != board.id && sp->is_child[x])
			{
				sp->update[x] = TRUE;
				smp_message(x, SMP_UPDATE, sp->id);
			}
	}
	/* Check for beta cutoffs separately, because while it may not be the best score anymore,
		it still requires us to clean up our local stuff. */
	if (sb->best_score >= sb->beta)
	{
		sp->no_moves_left = TRUE;

		LOCK(smp_data->lock);
		smp_data->stops_done++;
		UNLOCK(smp_data->lock);
		
		/* Now release the split point, detach, and go away. */
		UNLOCK(sp->lock);
		detach(sb);
	}
	UNLOCK(sp->lock);
#if 0
		SMP_DEBUG(print("cpu %i stopping siblings at %i\n", board.id, sp->id));
		for (x = 0; x < process_count; x++)
			if (x != board.id && sp->is_child[x])
			{
				sp->update[x] = TRUE;
				smp_message(x, SMP_STOP, 0);
			}
		UNLOCK(sp->lock);

		/* Wait for a message to arrive, and then (probably) unsplit. We do this because we sometimes
			get messages for other reasons, such as another processor gets a beta cutoff at the same time. */
		while (sp->active == TRUE && sp->child_count > 1)
		{
			while (smp_block[board.id].message_count == 0)
				;
			handle_messages(&sb);
		}
		/* This is so hacky, and I really hate having it here, but I can't think of a better solution.
			It's here because a processor could get a beta cutoff before any other processor attaches.
			It will bail out of the above loop because child_count == 1, but no processor will send
			an UNSPLIT message. Note that this unsplit could even fail because another processor got
			a beta cutoff too and told this processor to stop. Hey, nobody ever said DTS was easy. */
		if (sp->active == TRUE)
			sb = unsplit(sb, sp->id);
	}
	else
	{
		for (x = 0; x < process_count; x++)
			if (x != board.id && sp->is_child[x])
			{
				sp->update[x] = TRUE;
				smp_message(x, SMP_UPDATE, sp->id);
			}
		UNLOCK(sp->lock);
	}
#endif
}

/**
update():
This function is called when another processor has merged a value greater than alpha at a split point.
We need to check if we are doing unnecessary work, and update all the old bounds.
Created 082906; last modified 042007
**/
SEARCH_BLOCK *update(SEARCH_BLOCK *sb, int id)
{
	int side;
	int new_bound;
	SEARCH_BLOCK *new_sb;
	SPLIT_POINT **sp;

	/* Find the split point with the given id. */
	for (sp = board.split_point; *sp != NULL; sp--)
		if ((*sp)->id == id)
			break;
	if (*sp == NULL) /* the split point is no longer active */
		return sb;
	LOCK((*sp)->lock);
	/* Check if we've already updated here. This can happen if two processors get values > alpha
		one after another and then send two messages, when we only need to update once. */
	if ((*sp)->update[board.id] == FALSE)
	{
		UNLOCK((*sp)->lock);
		return sb;
	}
	(*sp)->update[board.id] = FALSE;
	/* From the search block of the split point, traverse down the tree and update all bounds. */
	new_sb = (*sp)->sb - (*sp)->board.search_stack + board.search_stack;
	new_bound = (*sp)->sb->alpha;
	UNLOCK((*sp)->lock);
	side = 0;
	for (; new_sb <= sb; new_sb++)
	{
		/* Check to see if we can narrow the window. If it's the other side to move, we negate the bound
			and check against beta, otherwise just check against alpha. */
		if (side == 0)
		{
			if (new_bound > new_sb->alpha)
				new_sb->alpha = new_bound;
		}
		else
		{
			if (-new_bound < new_sb->beta)
				new_sb->beta = -new_bound;
		}
		/* Somewhere down the tree, possibly the split point, we caused a beta cutoff, so back up a bit. */
		if (new_sb->alpha >= new_sb->beta)
		{
			new_sb->return_value = side == 0 ? new_bound : -new_bound;
			while (sb >= new_sb)
			{
				/* We must check for a complicated situation where there are one or more split points
					below a PV node that has been updated. The change in alpha can cause the other split
					points to have beta cutoffs. We don't send any messages in this situation, as all
					relevant processors will receive an UPDATE message from the PV node. We only detach
					ourselves from the split point, and in case we are the last child, unsplit it. */
				if (sb->ply == *board.split_ply)
				{
					SMP_DEBUG(print("cpu %i updating below split point\n", board.id));
					sb = detach(sb);
					/* Check if we didn't unsplit, then we can skip the tree updating and just stop. */
					if (sb->search_state == SEARCH_WAIT)
						return sb;
				}
				/* Check if we have made a move that needs to be unmade at this ply. */
				if (sb->move_made)
				{
					sb->move_made = FALSE;
					unmake_move();
				}
				sb--;
			}
			SMP_DEBUG(print("cpu %i updated, cutoff at %E\n%B", board.id, new_sb, &board));
			/* Continue searching at the node below the beta cutoff. */
			return new_sb - 1;
		}
		if (new_sb->search_state != SEARCH_IID) /* stm doesn't change on IID calls */
			side ^= 1;
	}
	return sb;
}

/**
insert():
This is called whenever a split point is created closer to the root than another split point along the
same path. Each processor needs to update both their internal split point stack and the stack on their
split point's internal board. This shouldn't happen too often because the first split will go as deep as
possible, and the node we split at now would usually be a better candidate.
Created 020208; last modified 022008
**/
void insert(BOARD *current_board, int id, SPLIT_POINT *new_sp)
{
	int *s_ply;
	SPLIT_POINT **sp;

	SMP_DEBUG(print("cpu %i trying to insert sp %i.\n", current_board->id, id));
	/* Find the split point. */
	for (sp = current_board->split_point; *sp != NULL; sp--)
		if ((*sp)->id == id)
			break;
	if (*sp == NULL)
	{
		SMP_DEBUG(print("insert bad: *sp=%p *(sp -1)=%p\n",*sp, *(sp-1)));
		return;
	}
	/* Now check which split point we should be inserting. If we are the splitting process, we already know it. */
	if (new_sp == NULL)
	{
		new_sp = *((*sp)->board.split_point - 1);
		if (new_sp == NULL)
		{
			SMP_DEBUG(print("sp_insert=NULL\n"));
			return;
		}
	}
	/* We've already inserted it... */
	if (*(sp - 1) != NULL && new_sp->id == (*(sp - 1))->id)
	{
		SMP_DEBUG(print("sp_->id=%i\n",new_sp->id));
		return;
	}
	SMP_DEBUG(print("cpu %i inserting sp %i.\n", current_board->id, id));

	/* Shift the split point stack forward and insert. */
	LOCK((*sp)->lock);
	for (sp = current_board->split_point, s_ply = current_board->split_ply; (*sp)->id != id; sp--, s_ply--)
	{
		*(sp + 1) = *sp;
		*(s_ply + 1) = *s_ply;
	}
	*(sp + 1) = *sp;
	*(s_ply + 1) = *s_ply;
	*sp = new_sp;
	*s_ply = new_sp->sb->ply;
	current_board->split_point++;
	current_board->split_ply++;
	UNLOCK((*sp)->lock);
	/* Now modify the boards inside the split points. */
	for (sp = current_board->split_point; (*sp)->id != id; sp--)
		insert(&(*sp)->board, id, new_sp);
}

/**
detach():
This function is called whenever a processor leaves a split point after finishing all its moves.
It checks if there is only one processor at the split point and if so tells it to unsplit. Otherwise
it detaches from every split point this processor is attached to and enters a waiting state.
Created 082906; last modified 041908
**/
SEARCH_BLOCK *detach(SEARCH_BLOCK *sb)
{
	int child_count;

	/* This block of code is used to prevent a deadlock that can occur when a processor joins the
		split point or a split point under it. */
	LOCK((*board.split_point)->lock);
	(*board.split_point)->no_moves_left = TRUE;
	child_count = (*board.split_point)->child_count;
	UNLOCK((*board.split_point)->lock);
	/* Now do the normal detaching. */
	if (child_count <= 1)
		sb = unsplit(sb, (*board.split_point)->id);
	else
		sb = stop(sb);
	return sb;
}

/**
stop():
This function is called when a processor needs to detach from every split point it is in. This is called in
three situations: Firstly, when the processor finishes searching its moves at a ply and there are still one or more
processors working at the split point; secondly, when another processor gets a beta cutoff at a split point we
are working on; and lastly, when the search is aborted and all child processors need to return to an idle state.
Created 082207; last modified 041908
**/
SEARCH_BLOCK *stop(SEARCH_BLOCK *sb)
{
	int x;
	SPLIT_POINT **sp;

	/* Go through and detach from each split point. */
	for (sp = board.split_point; *sp != NULL; sp--)
	{
		LOCK((*sp)->lock);
		SMP_DEBUG(print("cpu %i detaching from sp %i, %i children left\n", board.id, (*sp)->id, (*sp)->child_count));
		/* Set this now, as otherwise a processor could sneak in and attach after we unlock but before unsplitting. */
		(*sp)->no_moves_left = TRUE;
		/* Check if we are the last processor already. */
		if ((*sp)->child_count <= 1)
		{
			UNLOCK((*sp)->lock);
			sb = unsplit(sb, (*sp)->id);
			return sb;
		}
		/* Otherwise, remove this processor from the split point. */
		(*sp)->child_count--;
		(*sp)->is_child[board.id] = FALSE;
		/* Correct our local split point information. */
		board.split_point--;
		board.split_ply--;
		/* Now check if there is only one other processor. */
		if ((*sp)->child_count == 1)
		{
			for (x = 0; x < process_count; x++)
			{
				if ((*sp)->is_child[x])
				{
					SMP_DEBUG(print("cpu %i telling cpu %i to unsplit %i.\n", board.id, x, (*sp)->id));
					smp_message(x, SMP_UNSPLIT, (*sp)->id);
				}
			}
		}
		UNLOCK((*sp)->lock);
	}
	/* Go back to the root position to get ready for the next split point. */
	while (board.game_entry > root_entry)
		unmake_move();
	/* Tell the other processors we are idle. */
//	smp_block[board.id].idle = TRUE;
	smp_block[board.id].last_idle_time = get_time();
	sb->search_state = SEARCH_WAIT;
	return sb;
}

/**
unsplit():
If we are the last processor in a split point, then we need to destroy it. Also we copy the pv from
the split point into local memory if there is one, so that we may return it.
Created 082906; last modified 041908
**/
SEARCH_BLOCK *unsplit(SEARCH_BLOCK *sb, int id)
{
	SEARCH_BLOCK *new_sb;
	SPLIT_POINT **sp, **bottom;

	/* Find the split point with the given id. */
	for (bottom = board.split_point; *bottom != NULL; bottom--)
		if ((*bottom)->id == id)
			break;
	if (*bottom == NULL) /* the split point is no longer active */
		return sb;
	/* Update our local score if needed. */
	if ((*bottom)->update[board.id])
	{
/*		sb = update(sb, id);
		copy_pv(board.pv_stack[(*bottom)->sb->ply], (*bottom)->pv);
		if (sb->search_state == SEARCH_WAIT)
			return sb;
*/
		new_sb = sb;
		while (new_sb->ply > (*bottom)->sb->ply)
			new_sb--;
		if ((*bottom)->sb->alpha > new_sb->alpha)
		{
			new_sb->alpha = (*bottom)->sb->alpha;
			copy_pv(board.pv_stack[(*bottom)->sb->ply], (*bottom)->pv);
		}
	}

	/* Now go along split points, down the tree. We "recursively" lock the split point stack so that unsplitting
		is atomic with respect to each split point. */
	for (sp = board.split_point; sp >= bottom; sp--)
	{
		LOCK((*sp)->lock);
		SMP_DEBUG(print("cpu %i unsplitting sp %i\n", board.id, (*sp)->id));
		/* Another split point has attached to the split point since we got the message. While this cannot happen
			directly (a processor wouldn't attach to a split point with no moves left...), if there is a split point
			this one, another processor could attach to this one in order for it to "back up". */
		if ((*sp)->child_count > 1)
		{
			for (; sp <= board.split_point; sp++)
				UNLOCK((*sp)->lock);
			SMP_DEBUG(print("cpu %i failed unsplitting, sp %i has children.\n", board.id, (*sp)->id));
			return sb;
		}
	}
	/* Now that we have control over the split point stack, go through each one and unsplit it. */
	for (sp = bottom; sp <= board.split_point; sp++)
	{
		/* Clean up the split point... */
		(*sp)->is_child[board.id] = FALSE;
		(*sp)->child_count = 0;
		/* Now make sure the move list on this ply is empty. Trust me, this can be a big problem. */
		new_sb = sb;
		while (new_sb->ply > (*sp)->sb->ply)
			new_sb--;
		new_sb->next_move = new_sb->last_move;
		new_sb->select_state = SELECT_MOVE;

		UNLOCK((*sp)->lock);

		/* Update "global" information. */
		LOCK(smp_data->lock);
		(*sp)->active = FALSE;
		smp_data->split_count--;
		ASSERT(smp_data->split_count >= 0);
		UNLOCK(smp_data->lock);
	}

	/* Correct our local split point information. */
	for (sp = board.split_point; sp >= bottom; sp--)
	{
		board.split_point--;
		board.split_ply--;
		ASSERT(board.split_point >= board.split_point_stack);
		ASSERT(board.split_ply >= board.split_ply_stack);
	}
	return sb;
}

/**
copy_search_state():
This function copies the search state (the iterative stack) into shared memory when another processor
becomes idle and sends this processor a HELP message. The other processor can then find a suitable
split point on the stack, and tell this processor to split there, or just wait if no good split point is found.
Created 110907; last modified 110907
**/
void copy_search_state(SEARCH_BLOCK *sb, int process)
{
	SEARCH_BLOCK *sb_from;
	SEARCH_BLOCK *sb_to;
	TREE_BLOCK *tb;

	if (smp_block[board.id].idle)
		return;
	/* No need to set any locks, as only this processor will modify the certain entry. */
	tb = &tree_block[process * process_count + board.id];
	if (tb->used)
		return;
//	SMP_DEBUG(print("cpu %i copying tree state to cpu %i\n", board.id, process));
	sb_from = board.search_stack;
	sb_to = tb->search_stack;
	while (sb_from <= sb)
		*sb_to++ = *sb_from++;
	tb->search_block = sb_to;
	tb->used = TRUE;
}

/**
find_split_point():
After a HELP command is issued and one or more search states are copied, this function looks at each node in the
trees to see if any makes a good candidate split point. If so, it tells the processor to split there.
Created 110907; last modified 020208
**/
int find_split_point(void)
{
	float score;
	float fh_score;
	float depth_score;
	float moves_score;
	float best_score;
	int process;
	int best_process;
	SEARCH_BLOCK *sb;
	SEARCH_BLOCK *best_sb;
	TREE_BLOCK *tb;

	/* Iterate through all search stacks that might have been copied. */
//	(print("cpu %i examining tree states.\n", board.id));
	/* Make the best score be a minimum score for split points. */
//	best_score = MAX(smp_data->current_depth / 4, 14.);
best_score = 7.;
	best_process = -1;
	best_sb = NULL;

	for (process = 0; process < process_count; process++)
	{
		tb = &tree_block[board.id * process_count + process];
		if (tb->used)
		{
			for (sb = tb->search_stack; sb < tb->search_block; sb++)
			{
				/* Make sure this is a valid node to split at. */
				if (sb->search_state == QSEARCH_START || sb->search_state == QSEARCH_1 || sb->moves == 0)
					continue;

				if (sb->node_type == NODE_PV)
					fh_score = 1. - pow(.1, sb->moves - 2);
				else if (sb->node_type == NODE_ALL)
					fh_score = 1. - pow(.1, sb->moves - 1);
				else
					fh_score = 1. - pow(.1, sb->moves - 3);

/*				if (sb->node_type == NODE_PV)
					fh_score = 1. - (.1 * sb->moves - 2);
				else if (sb->node_type == NODE_ALL)
					fh_score = 1. - (.1 * sb->moves - 2);
				else
					fh_score = 1. - (.1 * sb->moves - 3);
*/
				depth_score = sb->depth;

				if (sb->select_state == SELECT_HASH_MOVE || sb->select_state == SELECT_THREAT_MOVE || sb->select_state == SELECT_GEN_MOVES)
					moves_score = .8;
				else
					moves_score = 1. - (.2 * (sb->last_move - sb->next_move - 2));

				/* Get a score based on the three factors. */
				score = fh_score * depth_score * moves_score;
//		(print("     %E s=%1.3f f=%1.3f d=%1.3f m=%1.3f\n", sb, score, fh_score, depth_score, moves_score));
				if (fh_score <= 0. || depth_score <= 0. || moves_score <= 0.)
					continue;

				if (score > best_score)
				{
					best_score = score;
					best_sb = sb;
					best_process = process;
				}
			}
			tb->used = FALSE;
		}
	}
	if (best_process != -1)
	{
//		(print("best %E s=%1.3f\n", best_sb, best_score));

		tb = &tree_block[board.id * process_count + best_process];
//		SMP_DEBUG(print("cpu %i telling cpu %i to split on %E.\n", board.id, best_process, best_sb));
//print("%1.2f\n",best_sb->depth / 4.);
		/* Tell the processor to split by ply and by search_block id, so that the correct split point is found. */
		return smp_tell(best_process, SMP_SPLIT, best_sb->ply | best_sb->id * MAX_PLY);
	}
	return -1;
}

#endif /* SMP */
